#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>

//void *pkcs7_parse_message(const void *data, size_t datalen);
//void *x509_cert_parse(const void *data, size_t datalen);
#define ARRAY_SIZE(x) (sizeof(x) / sizeof(*(x)))
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned long time64_t;

void kfree(void *ptr);

void up_write(void *sem) {}
void down_write(void *sem) {}
void up_read(void *sem) {}
void down_read(void *sem) {}
int _cond_resched() { return 0; }
void key_schedule_gc(time_t gc_at){}

int asn1_ber_decoder(const struct asn1_decoder *decoder,
		     void *context,
		     const unsigned char *data,
		     size_t datalen);

struct asymmetric_key_id {
	unsigned short	len;
	unsigned char	data[];
};

/* Shadow the definition from crypto/asymmetric_keys/x509_cert_parser.c */
struct x509_certificate *x509_cert_parse(const void *data, size_t datalen) {
  return NULL;
}

/* instead of linking in crypto/asymmetric_keys/asymmetric_type.c */
bool asymmetric_key_id_same(const struct asymmetric_key_id *kid1,
			    const struct asymmetric_key_id *kid2)
{
	if (!kid1 || !kid2)
		return false;
	if (kid1->len != kid2->len)
		return false;
	return memcmp(kid1->data, kid2->data, kid1->len) == 0;
}

struct public_key {
	void *key;
	u32 keylen;
	const char *id_type;
	const char *pkey_algo;
};
struct public_key_signature {
	struct asymmetric_key_id *auth_ids[2];
	u8 *s;			/* Signature */
	u32 s_size;		/* Number of bytes in signature */
	u8 *digest;
	u8 digest_size;		/* Number of bytes in digest */
	const char *pkey_algo;
	const char *hash_algo;
};

void public_key_signature_free(struct public_key_signature *sig)
{
	int i;

	if (sig) {
		for (i = 0; i < ARRAY_SIZE(sig->auth_ids); i++)
			kfree(sig->auth_ids[i]);
		kfree(sig->s);
		kfree(sig->digest);
		kfree(sig);
	}
}

void *kmemdup(const void *p, size_t size, int gfp) {
  void *ret = malloc(size);
  memcpy(ret, p, size);
  return ret;
}

time_t mktime64(const unsigned int year0, const unsigned int mon0,
		const unsigned int day, const unsigned int hour,
		const unsigned int min, const unsigned int sec) {
  struct tm tm;
  tm.tm_year = year0;
  tm.tm_mon = mon0;
  tm.tm_mday = day;
  tm.tm_hour = hour;
  tm.tm_min = min;
  tm.tm_sec = sec;
  return mktime(&tm);
}

struct asymmetric_key_id *asymmetric_key_generate_id(const void *val_1,
						     size_t len_1,
						     const void *val_2,
						     size_t len_2)
{
	struct asymmetric_key_id *kid;

	kid = malloc(sizeof(struct asymmetric_key_id) + len_1 + len_2);
	if (!kid)
		return (struct asymmetric_key_id*)-ENOMEM;
	kid->len = len_1 + len_2;
	memcpy(kid->data, val_1, len_1);
	memcpy(kid->data + len_1, val_2, len_2);
	return kid;
}

int printk(const char *s, ...) {
  va_list args;
  int ret;
  va_start(args, s);
  ret = vfprintf(stderr, s, args);
  va_end(args);
  return ret;
}

void *__kmalloc(size_t size, int flags) {
  return malloc(size);
}

void kfree(void *ptr) {
  free(ptr);
}

void kzfree(void *ptr) {
  // Let's hope nobody checks that the memory is zeroed.
  free(ptr);
}

void wait_for_completion(void *unused) { }
void wait_for_completion_killable(void *unused) { }
void wait_for_completion_killable_timeout(void *unused) { }
void init_waitqueue_head(void *unused) { }

void *__memcpy(void *dest, const void *src, size_t n) {
  return memcpy(dest, src, n);
}

extern const struct asn1_decoder x509_decoder;
///extern const struct asn1_decoder pkcs7_decoder;

struct x509_certificate {
	struct x509_certificate *next;
	struct x509_certificate *signer;	/* Certificate that signed this one */
	struct public_key *pub;			/* Public key details */
	struct public_key_signature *sig;	/* Signature parameters */
	char		*issuer;		/* Name of certificate issuer */
	char		*subject;		/* Name of certificate subject */
	struct asymmetric_key_id *id;		/* Issuer + Serial number */
	struct asymmetric_key_id *skid;		/* Subject + subjectKeyId (optional) */
	time64_t	valid_from;
	time64_t	valid_to;
	const void	*tbs;			/* Signed data */
	unsigned	tbs_size;		/* Size of signed data */
	unsigned	raw_sig_size;		/* Size of sigature */
	const void	*raw_sig;		/* Signature data */
	const void	*raw_serial;		/* Raw serial number in ASN.1 */
	unsigned	raw_serial_size;
	unsigned	raw_issuer_size;
	const void	*raw_issuer;		/* Raw issuer name in ASN.1 */
	const void	*raw_subject;		/* Raw subject name in ASN.1 */
	unsigned	raw_subject_size;
	unsigned	raw_skid_size;
	const void	*raw_skid;		/* Raw subjectKeyId in ASN.1 */
	unsigned	index;
	bool		seen;			/* Infinite recursion prevention */
	bool		verified;
	bool		self_signed;		/* T if self-signed (check unsupported_sig too) */
	bool		unsupported_key;	/* T if key uses unsupported crypto */
	bool		unsupported_sig;	/* T if signature uses unsupported crypto */
	bool		blacklisted;
};

struct x509_parse_context {
	struct x509_certificate	*cert;		/* Certificate being constructed */
	unsigned long	data;			/* Start of data */
	const void	*cert_start;		/* Start of cert content */
	const void	*key;			/* Key data */
	size_t		key_size;		/* Size of key data */
	int last_oid;
	int algo_oid;
///	enum OID	last_oid;		/* Last OID encountered */
///	enum OID	algo_oid;		/* Algorithm OID */
	unsigned char	nr_mpi;			/* Number of MPIs stored */
	u8		o_size;			/* Size of organizationName (O) */
	u8		cn_size;		/* Size of commonName (CN) */
	u8		email_size;		/* Size of emailAddress */
	u16		o_offset;		/* Offset of organizationName (O) */
	u16		cn_offset;		/* Offset of commonName (CN) */
	u16		email_offset;		/* Offset of emailAddress */
	unsigned	raw_akid_size;
	const void	*raw_akid;		/* Raw authorityKeyId in ASN.1 */
	const void	*akid_raw_issuer;	/* Raw directoryName in authorityKeyId */
	unsigned	akid_raw_issuer_size;
};

void public_key_free(struct public_key *key)
{
	if (key) {
		kfree(key->key);
		kfree(key);
	}
}

int LLVMFuzzerTestOneInput(const unsigned char *data,
                           unsigned long size) {
  struct x509_certificate *cert;
  struct x509_parse_context *ctx;
  cert = calloc(sizeof(struct x509_certificate), 1);
  cert->pub = calloc(sizeof(struct public_key), 1);
  cert->sig = calloc(sizeof(struct public_key_signature), 1);
  ctx = calloc(sizeof(struct x509_parse_context), 1);
  ctx->cert = cert;
  ctx->data = data;
  asn1_ber_decoder(&x509_decoder, ctx, data, size);
  free(ctx);
  free(cert->sig);
  free(cert->pub);
  free(cert);
  return 0;
}

#if 0
struct pkcs7_message {
	struct x509_certificate *certs;	/* Certificate list */
	struct x509_certificate *crl;	/* Revocation list */
	struct pkcs7_signed_info *signed_infos;
	u8		version;	/* Version of cert (1 -> PKCS#7 or CMS; 3 -> CMS) */
	bool		have_authattrs;	/* T if have authattrs */

	/* Content Data (or NULL) */
	///enum OID	data_type;	/* Type of Data */
	int data_type;
	size_t		data_len;	/* Length of Data */
	size_t		data_hdrlen;	/* Length of Data ASN.1 header */
	const void	*data;		/* Content Data (or 0) */
};

struct pkcs7_signed_info {
	struct pkcs7_signed_info *next;
	struct x509_certificate *signer; /* Signing certificate (in msg->certs) */
	unsigned	index;
	bool		unsupported_crypto;	/* T if not usable due to missing crypto */
	bool		blacklisted;

	/* Message digest - the digest of the Content Data (or NULL) */
	const void	*msgdigest;
	unsigned	msgdigest_len;

	/* Authenticated Attribute data (or NULL) */
	unsigned	authattrs_len;
	const void	*authattrs;
	unsigned long	aa_set;
#define	sinfo_has_content_type		0
#define	sinfo_has_signing_time		1
#define	sinfo_has_message_digest	2
#define sinfo_has_smime_caps		3
#define	sinfo_has_ms_opus_info		4
#define	sinfo_has_ms_statement_type	5
	time64_t	signing_time;

	/* Message signature.
 * 	 *
 * 	 	 * This contains the generated digest of _either_ the Content Data or
 * 	 	 	 * the Authenticated Attributes [RFC2315 9.3].  If the latter, one of
 * 	 	 	 	 * the attributes contains the digest of the the Content Data within
 * 	 	 	 	 	 * it.
 * 	 	 	 	 	 	 *
 * 	 	 	 	 	 	 	 * THis also contains the issuing cert serial number and issuer's name
 * 	 	 	 	 	 	 	 	 * [PKCS#7 or CMS ver 1] or issuing cert's SKID [CMS ver 3].
 * 	 	 	 	 	 	 	 	 	 */
	struct public_key_signature *sig;
};


struct pkcs7_parse_context {
	struct pkcs7_message	*msg;		/* Message being constructed */
	struct pkcs7_signed_info *sinfo;	/* SignedInfo being constructed */
	struct pkcs7_signed_info **ppsinfo;
	struct x509_certificate *certs;		/* Certificate cache */
	struct x509_certificate **ppcerts;
	unsigned long	data;			/* Start of data */
	///enum OID	last_oid;		/* Last OID encountered */
	int last_oid;
	unsigned	x509_index;
	unsigned	sinfo_index;
	const void	*raw_serial;
	unsigned	raw_serial_size;
	unsigned	raw_issuer_size;
	const void	*raw_issuer;
	const void	*raw_skid;
	unsigned	raw_skid_size;
	bool		expect_skid;
};


int LLVMFuzzerTestOneInput(const unsigned char *data,
                           unsigned long size) {
  struct pkcs7_parse_context *ctx;
  struct pkcs7_message *msg = (void*)-ENOMEM;
  ctx = calloc(sizeof(struct pkcs7_parse_context), 1);
  ctx->msg = calloc(sizeof(struct pkcs7_message), 1);
  ctx->sinfo = calloc(sizeof(struct pkcs7_signed_info), 1);
  ctx->sinfo->sig = calloc(sizeof(struct public_key_signature), 1);
  ctx->data = data;
  ctx->ppcerts = &ctx->certs;
  ctx->ppsinfo = &ctx->msg->signed_infos;
  asn1_ber_decoder(&pkcs7_decoder, ctx, data, size);
  free(ctx->sinfo->sig);
  free(ctx->sinfo);
  free(ctx->msg);
  free(ctx);
  return 0;
}
#endif
